@@ -16,7 +16,7 @@
     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.  '''
 
-__version__ = "0.3.04"
+__version__ = "0.3.05"
 __author__ = "nagev"
 __license__ = "GPLv3"
 
@@ -24,49 +24,94 @@ import re
 import sys
 import time
 import json
+import logging
 import urllib
 import urllib2
 from urlparse import parse_qs
 
-def _decrypt_signature(s):
-    # This function taken from youtube-dl source. Thanks!
-    if len(s) == 92:
-        return s[25] + s[3:25] + s[0] + s[26:42] + s[79] + s[43:79] + s[91] + \
-            s[80:83]
-    elif len(s) == 90:
-        return s[25] + s[3:25] + s[2] + s[26:40] + s[77] + s[41:77] + s[89] \
-            + s[78:81]
-    elif len(s) == 89:
-        return s[84:78:-1] + s[87] + s[77:60:-1] + s[0] + s[59:3:-1]
-    elif len(s) == 88:
-        return s[7:28] + s[87] + s[29:45] + s[55] + s[46:55] + s[2] +\
-            s[56:87] + s[28]
-    elif len(s) == 87:
-        return s[6:27] + s[4] + s[28:39] + s[27] + s[40:59] + s[2] + s[60:] 
-    elif len(s) == 86:
-        return s[83:36:-1] + s[0] + s[35:2:-1] 
-    elif len(s) == 85:
-        return s[83:34:-1] + s[0] + s[33:27:-1] + s[3] + s[26:19:-1] + s[34]\
-            + s[18:3:-1] + s[27]
-    elif len(s) == 84:
-        return s[81:36:-1] + s[0] + s[35:2:-1] 
-    elif len(s) == 83:
-        return s[81:64:-1] + s[82] + s[63:52:-1] + s[45] + s[51:45:-1] + s[1]\
-            + s[44:1:-1] + s[0]
-    elif len(s) == 82:
-        return s[1:19] + s[0] + s[20:68] + s[19] + s[69:82]
-    elif len(s) == 81:
-        return s[56] + s[79:56:-1] + s[41] + s[55:41:-1] + s[80] \
-            +s[40:34:-1] + s[0] + s[33:29:-1] + s[34] + s[28:9:-1] + s[29] +\
-            s[8:0:-1] + s[9]
-    elif len(s) == 80:
-        return s[1:19] + s[0] + s[20:68] + s[19] + s[69:80]
-    elif len(s) == 79:
-        return s[54] + s[77:54:-1] + s[39] + s[53:39:-1] + s[78] + \
-            s[38:34:-1] + s[0] + s[33:29:-1] + s[34] + s[28:9:-1] + \
-            s[29] + s[8:0:-1] + s[9]
+logging.basicConfig(level=logging.INFO)
+
+def dojsfunc2(f1argname, f2arg1, f2arg2, f2name, f2bod, sig):
+    f2parts = f2bod.split(";")
+    newvarname = ""
+    newvarval = 0
+    for part in f2parts:
+        # newvar is an index of sig
+        if re.match(r'var\s(\w)=(\w)\[(\d+)\]', part):
+            match = re.match(r'var\s(\w)=(\w)\[(\d+)\]', part)
+            newvarname = match.group(1)
+            if match.group(2) == f1argname:
+                newvarval = sig[int(match.group(3))]
+            else:
+                raise RuntimeError("no match in f2 for part: %s" % part)
+        # a[n]=a[b%a.length]
+        elif re.match(r'(\w)\[(\d+)\]=(\w)\[(\w)\%(\w)\.length\]', part):
+            match = re.match(r'(\w)\[(\d+)\]=(\w)\[(\w)\%(\w)\.length\]', part)
+            if match.group(1) == f1argname and match.group(3) == f1argname and\
+                match.group(4) == 'b' and match.group(5) == f1argname:
+                index = int(match.group(2))
+                newchar = sig[(int(f2arg2) % len(sig))]
+                sig = sig[:index] + newchar + sig[index + 1:]
+            else:
+                raise RuntimeError("no match in f2 for part: %s" % part)
+        # a[b]=c
+        elif re.match(r'(\w)\[(\w)\]=(\w)', part):
+            match = re.match(r'(\w)\[(\w)\]=(\w)', part)
+            if match.group(1) == f1argname and match.group(2) == 'b' and\
+                match.group(3) == newvarname:
+                index = int(f2arg2)
+                sig = sig[:index] + str(newvarval) + sig[index + 1:]
+            else:
+                raise RuntimeError("no match in f2 for part: %s" % part)
+        elif re.match(r'return\sa', part):
+            return sig
+        else:
+            raise RuntimeError("no match in f2 for part: %s" % part)
+            
+def dojs(sig, f1arg, f1bod, f2name=None, f2arg=None, f2bod=None):
+    # split js function components
+    f1parts = f1bod.split(";")
+    for part in f1parts:
+        # split, do nothing
+        if re.match(r'%s=%s\.split\(""\)' % (f1arg, f1arg), part):
+            pass
+        # call secondary function
+        elif re.match(r'%s=%s\((\w+),(\w+)\)' % (f1arg, f2name), part):
+            match = re.match(r'%s=%s\((\w+),(\w+)\)' % (f1arg, f2name), part)
+            f2arg1, f2arg2 = match.group(1), match.group(2)
+            sig = dojsfunc2(f1arg, f2arg1, f2arg2, f2name, f2bod, sig)
+        # reverse        
+        elif re.match(r'%s=%s\.reverse\(\)' % (f1arg, f1arg), part):
+            sig = sig[::-1]
+        # slice
+        elif re.match(r'%s=%s\.slice\((\d+)\)' % (f1arg, f1arg), part):
+            match = re.match(r'%s=%s\.slice\((\d+)\)' % (f1arg, f1arg), part)
+            sliceval = int(match.group(1))
+            sig = sig[sliceval:]
+        # return
+        elif re.match(r'return %s\.join\(""\)' % f1arg, part):
+            return sig
+        else:
+            raise RuntimeError("no match for %s" % part)
+
+def _decodesig(s, js):
+    logging.debug("sig length: %s" % len(s))
+    match = re.search(r'g.sig\|\|(\w+)\(g.s\)', js)
+    f1name = match.group(1)
+    match = re.search(r'function %s\((\w+)\)\{([^\{]+)\}' % f1name, js)
+    f1arg = match.group(1)
+    f1bod = match.group(2)
+    match = re.search(r'(\w+)\(\w+,\d+\)', f1bod)
+    if match:
+        f2name = match.group(1)
+        match = re.search(r'(function %s\((\w+,\w+)\)\{([^\{]+)\})' % f2name,
+               js)
+        f2 = match.group(1)
+        f2arg = match.group(2)
+        f2bod = match.group(3)
+        return dojs(s, f1arg, f1bod, f2name, f2arg, f2bod)
     else:
-        raise NameError("Unable to decode video url - sig len %s" % len(s))
+        return dojs(s, f1arg, f1bod)
 
 class Stream():
     resolutions = {
@@ -88,9 +133,11 @@ class Stream():
         '100': ('640x360-3D', 'webm'),
         '102': ('1280x720-3D', 'webm')}
 
-    def __init__(self, streammap, opener, title="ytvid"):
+    def __init__(self, streammap, opener, title="ytvid", js=None):
         if not streammap.get("sig", ""):
-            streammap['sig'] = [_decrypt_signature(streammap['s'][0])]
+            logging.debug("Decrypting sig: %s" % streammap['s'])
+            streammap['sig'] = [_decodesig(streammap['s'][0], js)]
+            logging.debug("Calculated decrypted sig: %s" % streammap['sig'][0])
         self.url = streammap['url'][0] + '&signature=' + streammap['sig'][0]
         self.vidformat = streammap['type'][0].split(';')[0]
         self.resolution = self.resolutions[streammap['itag'][0]][0]
@@ -107,7 +154,8 @@ class Stream():
     def download(self, progress=True, filepath=""):
         response = self._opener.open(self.url)
         total = int(response.info().getheader('Content-Length').strip())
-        print u"-Downloading '{}' [{:,} Bytes]".format(self.filename, total)
+        print (u"-Downloading '{}' [{:,} Bytes]".format(self.filename,
+            total)).encode('UTF-8')
         status_string = ('  {:,} Bytes [{:.2%}] received. Rate: [{:4.0f} '
                          'kbps].  ETA: [{:.0f} secs]')
         chunksize, bytesdone, t0 = 1024, 0, time.time()
@@ -169,7 +217,9 @@ class Pafy():
               "Trident/5.0)")
         opener.addheaders = [('User-Agent', ua)]
         self.keywords = ""
+        logging.debug("requested info page: %s" % infoUrl)
         self.rawinfo = opener.open(infoUrl).read()
+        logging.debug("got info")
         self.allinfo = parse_qs(self.rawinfo)
         self.title = self.allinfo['title'][0].decode('utf-8')
         self.author = self.allinfo['author'][0]
@@ -189,9 +239,12 @@ class Pafy():
             self.bigthumbhd = self.allinfo['iurlmaxres'][0]
         streamMap = self.allinfo['url_encoded_fmt_stream_map'][0].split(',')
         smap = [parse_qs(sm) for sm in streamMap]
+        js = None
         if not smap[0].get("sig", ""):  # vevo!
             watchurl = "https://www.youtube.com/watch?v=" + vidid
+            logging.debug("request watch?v page: %s" % watchurl)
             watchinfo = opener.open(watchurl).read()
+            logging.debug("got watch?v page")
             match = re.search(r';ytplayer.config = ({.*?});', watchinfo)
             try:
                 myjson = json.loads(match.group(1))
@@ -199,8 +252,12 @@ class Pafy():
                 raise NameError('Problem handling this video')
             args = myjson['args']
             streamMap = args['url_encoded_fmt_stream_map'].split(",")
+            html5player = myjson['assets']['js']
+            logging.debug("getting js url: %s" % html5player)
+            js = opener.open(html5player).read()
+            logging.debug("got js from %s" % html5player)
             smap = [parse_qs(sm) for sm in streamMap]
-        self.streams = [Stream(sm, opener, self.title) for sm in smap]
+        self.streams = [Stream(sm, opener, self.title, js) for sm in smap]
 
     def getbest(self, preftype="any", ftypestrict=True):
         # set ftypestrict to False to use a non preferred format if that
